//Sorta Sudoku Guessing Game
//Written by Boomer48: December 2019

#include "LCD_2004.h"
#include "Keypad_4x4.h"
#include <EEPROM.h>

char Game_Array[4][4] = {
  {' ', ' ', ' ', ' '},
  {' ', ' ', ' ', ' '},
  {' ', ' ', ' ', ' '},
  {' ', ' ', ' ', ' '}
};

const char Line_Array[24][4] = {
  {'1', '2', '3', '4'},
  {'1', '2', '4', '3'},
  {'1', '3', '2', '4'},
  {'1', '3', '4', '2'},
  {'1', '4', '2', '3'},
  {'1', '4', '3', '2'},
  {'2', '1', '3', '4'},
  {'2', '1', '4', '3'},
  {'2', '3', '1', '4'},
  {'2', '3', '4', '1'},
  {'2', '4', '1', '3'},
  {'2', '4', '3', '1'},
  {'3', '2', '1', '4'},
  {'3', '2', '4', '1'},
  {'3', '1', '2', '4'},
  {'3', '1', '4', '2'},
  {'3', '4', '2', '1'},
  {'3', '4', '1', '2'},
  {'4', '2', '3', '1'},
  {'4', '2', '1', '3'},
  {'4', '3', '2', '1'},
  {'4', '3', '1', '2'},
  {'4', '1', '2', '3'},
  {'4', '1', '3', '2'}
};

//LCD addresses of 4x4 matrix
byte LCD_Addr[16] = {0x88, 0x89, 0x8A, 0x8B, 0xC8, 0xC9, 0xCA, 0xCB, 0x9C, 0x9D, 0x9E, 0x9F, 0xDC, 0xDD, 0xDE, 0xDF};

char Display_Vals[] = "                ";

byte Line_Num;
byte ROM_Byte;
byte attempts;
char ASCII[] = "0123456789ABCDEF";

void setup()
{
  Keypad_setup();
  LCD_setup();
  Gen_Puzzle();
}

void loop()
{
  byte key1 = No_Key;
  byte key2 = No_Key;
  byte key_count = 0;
  byte ASC_Char;
  byte timeout = 0;

  // Each key press increments the displayed value
  // Exit after no activity for 1 second
  while (timeout < 20) {
    key1 = Get_Key();

    if (key2 == No_Key)
      key2 = key1;

    if (key1 != key2)
      key1 = No_Key;

    if ((key1 != No_Key) && (Display_Vals[key1] == ' '))
    {
      if (key_count == 4)
        key_count = 1;
      else
        key_count++;

      ASC_Char = key_count + '0';
      LCD_Char_Loc(LCD_Addr[key1], ASC_Char);
      timeout = 0;
    }
    else
    {
      delay(50);
      timeout++;
    }
  }

  if ((key2 != No_Key) && (Display_Vals[key2] == ' '))
  {
    if (ASC_Char == Game_Array[key2 >> 2][key2 % 4])
      Display_Vals[key2] = ASC_Char;
    else
      LCD_Char_Loc(LCD_Addr[key2], ' ');

    attempts++;
    if ((attempts & 0x0F) % 10 == 0)
      attempts = attempts + 6; //BCD adjustment
    LCD_Char_Loc(0x92, (ASCII[attempts >> 4]));
    LCD_Char_Loc(0x93, (ASCII[attempts & 0xF]));
  }
}

// Generate a 4x4 puzzle
void Gen_Puzzle() {
  byte i = 0;
  byte j = 0;
  char junk;

  LCD_Line1();
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte('G');
  Send_Byte('E');
  Send_Byte('N');
  Send_Byte('E');
  Send_Byte('R');
  Send_Byte('A');
  Send_Byte('T');
  Send_Byte('I');
  Send_Byte('N');
  Send_Byte('G');
  LCD_Line2();
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte('P');
  Send_Byte('U');
  Send_Byte('Z');
  Send_Byte('Z');
  Send_Byte('L');
  Send_Byte('E');

  delay(1000);

  // reset array for a new game
  while (i < 4) {
    while (j < 4) {
      Game_Array[i][j] = ' ';
      j++;
    }
    i++;
    j = 0;
  }

  i = 0;
  j = 0;

  ROM_Byte = EEPROM.read(0);
  randomSeed(ROM_Byte);
  ROM_Byte = random(256);
  EEPROM.write(0, ROM_Byte);

  // randomly select first three lines
  while (i < 3) {
    Line_Num = byte(random(2400) % 24);

    if ((Game_Array[0][0] != Line_Array[Line_Num][0]) && (Game_Array[1][0] != Line_Array[Line_Num][0]) &&
        (Game_Array[0][1] != Line_Array[Line_Num][1]) && (Game_Array[1][1] != Line_Array[Line_Num][1]) &&
        (Game_Array[0][2] != Line_Array[Line_Num][2]) && (Game_Array[1][2] != Line_Array[Line_Num][2]) &&
        (Game_Array[0][3] != Line_Array[Line_Num][3]) && (Game_Array[1][3] != Line_Array[Line_Num][3]))
    {
      Game_Array[i][0] = Line_Array[Line_Num][0];
      Game_Array[i][1] = Line_Array[Line_Num][1];
      Game_Array[i][2] = Line_Array[Line_Num][2];
      Game_Array[i][3] = Line_Array[Line_Num][3];
      i++;
    }
  }

  i = 0;
  j = 1;

  // fill the last line (only one solution is available)
  while (i < 4) {
    while (j <= 4) {
      junk = ASCII[j];
      if ((Game_Array[0][i] != junk) && (Game_Array[1][i] != junk) && (Game_Array[2][i] != junk))
      {
        Game_Array[3][i] = junk;
        j = 5;
      }
      else
        j++;
    }
    i++;
    j = 1;
  }

  i = 0;
  j = 0;

  delay(1000);
  Clear_LCD();
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte('R');
  Send_Byte('E');
  Send_Byte('A');
  Send_Byte('D');
  Send_Byte('Y');
  delay(1000);
  Clear_LCD();

  //Display first digit
  junk = Game_Array[0][0];
  Display_Vals[0] = junk;
  LCD_Char_Loc(LCD_Addr[0], junk);
  LCD_Char_Loc(LCD_Addr[0] - 1, '(');
  LCD_Char_Loc(LCD_Addr[3] + 1, ')');
  LCD_Char_Loc(LCD_Addr[4] - 1, '(');
  LCD_Char_Loc(LCD_Addr[7] + 1, ')');
  LCD_Char_Loc(LCD_Addr[8] - 1, '(');
  LCD_Char_Loc(LCD_Addr[11] + 1, ')');
  LCD_Char_Loc(LCD_Addr[12] - 1, '(');
  LCD_Char_Loc(LCD_Addr[15] + 1, ')');
  attempts = 0;
  LCD_Char_Loc(0x92, (ASCII[attempts >> 4]));
  LCD_Char_Loc(0x93, (ASCII[attempts & 0xF]));
}

byte Get_Key()
{
  byte key1;
  byte key2;

  key1 = Scan_Keypad();

  if (key1 != No_Key)
  {
    delay(50); // debounce
    key2 = No_Key - 1;

    while (key2 != No_Key)
    { // wait for key release
      key2 = Scan_Keypad();
      delay(50); // debounce
    }
  }
  return key1;
}
